#include <protocol/protocol.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <elib/timer.h>
#include <elib/machine.h>
#include <unix.h>

#define BIND_RETRY_INTERVAL 5000
#define CONNECT_RETRY_INTERVAL 5000

typedef struct tconnection {
  int fd;
  io_handler i;
  layer l;
  int addrsize;
  struct sockaddr_in *addr;
  int connected;
  io_handler connect_handler;
} *tconnection;


typedef struct tserver {
  int fd;
  int addrsize;
  struct sockaddr_in *addr;
  void (*new_client_handler)();
  void *new_client_handler_arg;
} *tserver;


#ifdef sun
#include <sys/filio.h>
#else
#include <sys/ioctl.h>
#endif

#include <sys/errno.h>
extern int errno;
struct sockaddr_in *make_sockaddr_in();

#ifdef sun
#define ADDR_SUFFIX S_un.S_addr
#else
#define ADDR_SUFFIX s_addr
#endif

static int tcp_shutdown(tconnection t)
{
  remove_io_handler(t->i);
  deallocate_memory(t->l);
/*  deallocate_memory(t->addr); */  /* Not ours to deallocate */
  deallocate_memory(t);
  close(t->fd);
}

static int tcp_read(tconnection t)
{
  int bytes_available=0;
  pbuffer b;

  if (ioctl(t->fd,FIONREAD,&bytes_available)==0){
    if (bytes_available > 0) {
      b=allocate_buffer(bytes_available);
      read(t->fd,b->contents,bytes_available);
      write_up(t->l,b,0);
      dereference_buffer(b);
      return(1);
    }
  }
  write_up(t->l,0,PROTOCOL_STACK_SHUTDOWN);
  tcp_shutdown(t);
  return(0);
}

static int tcp_write(layer l, pbuffer b, int f)
{
  int offset=0;
  tconnection t=(tconnection)(l->state);
  if (b) write(t->fd,b->contents,b->length);
  if (f & PROTOCOL_STACK_SHUTDOWN) tcp_shutdown(t);
}

static void from_below(){} /* trap connected propagation*/

static int connect_finish(tconnection t)
{
  int blocking=0;
  t->connected=1;
  ioctl(t->fd,FIONBIO,&blocking);
  write_up(t->l,0,PROTOCOL_CONNECTED);
  t->i=register_io_handler(tcp_read,t->fd,READ_HANDLER,t);
  return(0);
}

static int connect_try (tconnection t)
{
  struct sockaddr_in *a=(struct sockaddr_in *)t->addr;
  int temp;
  int blocking=0;
  if (t->connected) return(0);
  if (t->fd){
    remove_timer(t->connect_handler);
    close(t->fd);
  }
  t->fd=socket(AF_INET,SOCK_STREAM,0);
  ioctl(t->fd,FIONBIO,&blocking);
  if (connect (t->fd,(void *)t->addr,t->addrsize) == 0){
    return(connect_finish(t));
  }
  blocking=0;
  ioctl(t->fd,FIONBIO,&blocking);
  if (errno==EINPROGRESS)
    t->connect_handler= 
      register_io_handler(connect_finish,t->fd,WRITE_HANDLER,t);
  return(1);
}


static int new_connection(tserver t)
{
  layer l;
  tconnection new=(tconnection)allocate_memory(sizeof(struct tconnection));
  struct sockaddr_in from;
  int addrsize=sizeof(struct sockaddr_in);

  if ((new->fd=accept(t->fd,(void *)&from,&addrsize)) > 0){
    new->i=register_io_handler(tcp_read,new->fd,
			       READ_HANDLER|EXCEPTION_HANDLER,new);
    (*t->new_client_handler)(new->l=create_layer(new,tcp_write,0,0),
			     ntohl(from.sin_addr.ADDR_SUFFIX),
			     t->new_client_handler_arg);
    return(1);
  }
  deallocate_memory(new);
  return(1);
}


static int bind_try(tserver t)
{
  struct sockaddr_in *a=t->addr;
  unsigned short z=vhtons(a->sin_port);
  unsigned int k=z;
  if (bind (t->fd,(void *)t->addr,t->addrsize) == 0){
    listen(t->fd,5);
    printf ("bound: %d\n",k);
    register_io_handler(new_connection,t->fd,READ_HANDLER,t);
    return(0);
  }
  return(1);
}


layer tcp_create_client (struct sockaddr_in *addr)
{
  tconnection new=(tconnection)allocate_memory(sizeof(struct tconnection));
  new->addr=addr;
  new->addrsize=sizeof(struct sockaddr_in);
  new->fd=0;
  new->connected=0;
  register_timer(CONNECT_RETRY_INTERVAL,connect_try,new);
  return(new->l=create_layer(new,tcp_write,from_below,0));
}


void tcp_create_server(struct sockaddr_in *addr,
		       void (*new_client_handler)(),
		       void *new_client_handler_arg)
{
  tserver new=(tserver)allocate_memory(sizeof(struct tserver));
  new->addr=addr;
  new->addrsize=sizeof(struct sockaddr_in);
  new->new_client_handler=new_client_handler;
  new->new_client_handler_arg=new_client_handler_arg;
  new->fd=socket(AF_INET,SOCK_STREAM,0);
  register_timer(BIND_RETRY_INTERVAL,bind_try,new);
}
